/*
** scansbr.c -- routines to help scan along...
**
** This code is Copyright (c) 2002, by the authors of nmh.  See the
** COPYRIGHT file in the root directory of the nmh distribution for
** complete copyright information.
*/

#include <h/mh.h>
#include <h/addrsbr.h>
#include <h/fmt_scan.h>
#include <h/scansbr.h>
#include <h/tws.h>
#include <h/utils.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sysexits.h>

#ifdef _FSTDIO
# define _ptr _p  /* Gag */
# define _cnt _w  /* Wretch */
#endif

#define MAXSCANL 256  /* longest possible scan line */

/*
** Buffer size for content part of header fields.  We want this
** to be large enough so that we don't do a lot of extra FLDPLUS
** calls on m_getfld.
*/
#define SBUFSIZ 512

static struct format *fmt;

static struct comp *datecomp;  /* pntr to "date" comp */
static int ncomps = 0;  /* # of interesting components */
static char **compbuffers = NULL;  /* buffers for component text */
static struct comp **used_buf = NULL;  /* stack for comp that use buffers */

static int dat[5];  /* aux. data for format routine */

char *scanl = NULL;  /* text of most recent scanline */

#define FPUTS(buf) {\
		if (fputs(buf, scnout) == EOF)\
			adios(EX_IOERR, scnmsg, "write error on");\
	}

/*
** prototypes
*/
int sc_width(void);  /* from termsbr.c */

#ifdef MULTIBYTE_SUPPORT
#define SCAN_CHARWIDTH MB_CUR_MAX
#else
#define SCAN_CHARWIDTH 1
#endif

int
scan(FILE *inb, int innum, int outnum, char *fmtstr, int width, int curflg,
	int unseen)
{
	static int slwidth;
	int i, compnum, state;
	unsigned char *cp, *tmpbuf;
	char **nxtbuf;
	struct comp *cptr;
	struct comp **savecomp;
	char *scnmsg = NULL;
	FILE *scnout = NULL;
	char name[NAMESZ];
	int incing = (outnum != SCN_MBOX && outnum != SCN_FOLD);
	int scanfolder = (outnum == SCN_FOLD);
	long fpos;
	struct stat st;

	/* first-time only initialization */
	if (!scanl) {
		if (incing)
			umask(~m_gmprot());

		if (fmtstr) {
			if (width == 0) {
				if ((width = sc_width()) < WIDTH/2)
					width = WIDTH/2;
				else if (width > MAXSCANL)
					width = MAXSCANL;
			}
			dat[3] = slwidth = width;
			scanl = (char *) mh_xmalloc((size_t) SCAN_CHARWIDTH *
					(slwidth + 2));  /* probably for \n and \0 */
			/* Compile format string */
			ncomps = fmt_compile(fmtstr, &fmt) + 1;
			FINDCOMP(datecomp, "date");
		} else {
			ncomps = 1;
			datecomp = NULL;
		}

		nxtbuf = compbuffers = (char **) mh_xcalloc((size_t) ncomps,
				sizeof(char *));
		used_buf = (struct comp **) mh_xcalloc((size_t) (ncomps+1),
				sizeof(struct comp *));
		/* NULL-terminate array */
		used_buf += ncomps;
		*used_buf = NULL;
		/* allocate space for the items */
		for (i = ncomps; i--; )
			*nxtbuf++ = mh_xmalloc(SBUFSIZ);
	}

	/*
	** each-message initialization
	*/
	nxtbuf = compbuffers;
	savecomp = used_buf;
	tmpbuf = *nxtbuf++;
	dat[0] = innum ? innum : outnum;
	dat[1] = curflg;
	dat[4] = unseen;
	fpos = ftell(inb);

	/*
	** Get the first field.  If the message is non-empty
	** and we're doing an "inc", open the output file.
	*/
	if ((state = m_getfld(FLD, name, tmpbuf, SBUFSIZ, inb)) == FILEEOF) {
		if (ferror(inb)) {
			advise("read", "unable to"); /* "read error" */
			return SCNFAT;
		} else {
			return SCNEOF;
		}
	}

	if (incing) {
		scnmsg = m_name(outnum);
		if (*scnmsg == '?')  /* msg num out of range */
			return SCNNUM;
		if (!(scnout = fopen(scnmsg, "w")))
			adios(EX_IOERR, scnmsg, "unable to write");
	}

	/* scan - main loop */
	for (compnum = 1; ;
			state = m_getfld(state, name, tmpbuf, SBUFSIZ, inb)) {
		switch (state) {
		case FLD:
		case FLDPLUS:
			compnum++;
			if (incing) {
				FPUTS(name);
				FPUTS(":");
				FPUTS(tmpbuf);
			}
			/*
			** if we're interested in this component, save
			** a pointer to the component text, then start
			** using our next free buffer as the component
			** temp buffer (buffer switching saves an extra
			** copy of the component text).
			*/
			if (fmtstr && (cptr = wantcomp[CHASH(name)])) {
				do {
					if (mh_strcasecmp(name, cptr->c_name)!=0) {
						continue;
					}
					if (!cptr->c_text) {
						cptr->c_text = tmpbuf;
						cp = tmpbuf+strlen(tmpbuf)-1;
						for (; cp >= tmpbuf; cp--) {
							if (isspace(*cp))
								*cp = '\0';
							else
								break;
						}
						*--savecomp = cptr;
						tmpbuf = *nxtbuf++;
					}
					break;
				} while ((cptr = cptr->c_next));
			}

			while (state == FLDPLUS) {
				state = m_getfld(state, name, tmpbuf, SBUFSIZ,
						inb);
				if (incing)
					FPUTS(tmpbuf);
			}
			break;

		case BODY:
			compnum = -1;
			if (scanfolder) {
				/* stop here if we scan a msg in a folder */
				state = FILEEOF;
				goto finished;
			}
			/* otherwise (mbox): snarf the body */
			if (incing) {
				FPUTS("\n");
				FPUTS(tmpbuf);
			}
body:;
			while (state == BODY) {
				state = m_getfld(state, name, tmpbuf, SBUFSIZ,
						inb);
				if (incing) {
					FPUTS(tmpbuf);
				}
			}
			goto finished;

		case LENERR:
		case FMTERR:
			fprintf(stderr, innum ? "??Format error (message %d) in " : "??Format error in ", outnum ? outnum : innum);
			fprintf(stderr, "component %d\n", compnum);

			if (incing) {
				FPUTS("\n\nBAD MSG:\n");
				FPUTS(name);
				FPUTS("\n");
				state = BODY;
				goto body;
			}
			/* fall through if we scan only */

		case FILEEOF:
			goto finished;

		default:
			adios(EX_SOFTWARE, NULL, "getfld() returned %d", state);
		}
	}

finished:

	/* Format and output the scan line. */
	if (ferror(inb)) {
		advise("read", "unable to");
		return SCNFAT;
	}

	if (incing) {
		if ((dat[2] = ftell(scnout)) == EOF)
			adios(EX_IOERR, scnmsg, "write error on");
	} else if (!scanfolder) {
		if ((dat[2] = ftell(inb)) == EOF)
			adios(EX_IOERR, scnmsg, "write error on");
		dat[2] -= fpos;
	}

	if (fmtstr) {
		fstat(fileno(inb), &st);
		if (scanfolder) {
			dat[2] = st.st_size;
		}
		if (datecomp && !datecomp->c_text) {
			if (!datecomp->c_text) {
				if (!datecomp->c_tws)
					datecomp->c_tws = (struct tws *) mh_xcalloc((size_t) 1, sizeof(*datecomp->c_tws));
				if (!datecomp->c_tws)
					adios(EX_OSERR, NULL, "unable to allocate tws buffer");
				*datecomp->c_tws = *dlocaltime((time_t *) &st.st_mtime);
				datecomp->c_flags |= CF_DATEFAB|CF_TRUE;
			} else {
				datecomp->c_flags &= ~CF_DATEFAB;
			}
		}

		/* print the scan line */
		fmt_scan(fmt, scanl, slwidth, dat);
		fputs(scanl, stdout);
	}

	/* return dynamically allocated buffers to pool */
	while ((cptr = *savecomp++)) {
		*--nxtbuf = cptr->c_text;
		cptr->c_text = NULL;
	}
	*--nxtbuf = tmpbuf;

	if (incing && (ferror(scnout) || fclose(scnout) == EOF))
		adios(EX_IOERR, scnmsg, "write error on");

	return (state != FILEEOF ? SCNERR : SCNMSG);
}
